#include "Util-pkg.h"
#include "P3d-pkg.h"
#include "Planner-pkg.h"
#include "Localpath-pkg.h"
#include "Collision-pkg.h"
#include "Graphic-pkg.h"

#ifdef LIGHT_PLANNER
#include "LightPlanner-pkg.h"
#endif

/*Warning : This variable is also defined in FORMOptim for elastic smoothing algorithm. This one concerns only random algorithm.*/
#define MAX_NB_TRY_OPTIM  20

int QUICK_DESCENT = TRUE;
double EPS_ELASTIC = 0.0001;
double D0 = 0;

int STOP_OPTIM = FALSE;

/*
 *  Extract the local path that contains the configuration
 *  at a given length along a trajectory
 *
 *  In: The trajectory, The length
 *  Out: The local path, the distance along the local path of the
 *       configuration from the beginning of the local path
 */

p3d_localpath *p3d_localpath_from_traj(p3d_localpath *localpathPt,
                                       double length,
                                       double *distlocPt) {
  double l = 0;

  l = localpathPt->length_lp;

  while ((l < length) && (localpathPt != NULL)) {
    localpathPt = localpathPt->next_lp;
    l += localpathPt->length_lp;
  }

  l -= localpathPt->length_lp;
  *distlocPt = length - l;
  return localpathPt;
}

void p3d_remove_useless_nodes(p3d_traj* traj){

    p3d_localpath* lp_cur=traj->courbePt;

    int n;

        lp_cur=traj->courbePt->next_lp;
        while(lp_cur!=NULL){
            p3d_localpath* lp_prev=lp_cur->prev_lp;
            configPt q_init=lp_prev->config_at_param(traj->rob,lp_prev,0);
            configPt q_end=lp_cur->config_at_param(traj->rob,lp_cur,lp_cur->length_lp);
            p3d_localpath* lp_test=p3d_local_planner(traj->rob,q_init,q_end);
            if((fabs(lp_test->length_lp-(lp_prev->length_lp+lp_cur->length_lp))<1e-5 ||
                lp_test->length_lp-(lp_prev->length_lp+lp_cur->length_lp)<=0) &&
                !p3d_unvalid_localpath_test(traj->rob,lp_test,&n)){
                lp_test->prev_lp=lp_prev->prev_lp;
                lp_test->next_lp=lp_cur->next_lp;
                MY_FREE(lp_cur,p3d_localpath,1);
                MY_FREE(lp_prev,p3d_localpath,1);
                lp_cur=lp_test;
                if(lp_cur->next_lp!=NULL){
                    lp_cur->next_lp->prev_lp=lp_cur;
                }
                if(lp_cur->prev_lp==NULL){
                    traj->courbePt=lp_cur;
                }
                else{
                    lp_cur->prev_lp->next_lp=lp_cur;
                }
                traj->nlp--;
            }
            else{
                MY_FREE(lp_test,p3d_localpath,1);
            }
            lp_cur=lp_cur->next_lp;
        }

        lp_cur=traj->courbePt;
        while(lp_cur->next_lp!=NULL){
            lp_cur=lp_cur->next_lp;
        }

    lp_cur=traj->courbePt;
    traj->range_param=0;
    while(lp_cur!=NULL){
        traj->range_param=traj->range_param+lp_cur->length_lp;
        lp_cur=lp_cur->next_lp;
    }

}

int p3d_optimize_traj(p3d_rob* robot, p3d_traj* traj, int rand, int elastic, int clean, void (*fct_draw)(void)){
  int i, ntest = 0, nb_optim = 0;
  double gain = 1., gaintot = 1., epsilon = 0.0;

  STOP_OPTIM = FALSE;
#ifdef DPG
  if (traj) {
    traj->isOptimized = true;
  }
#endif
   if (!traj || traj->nlp <= 1) {
      if (ENV.getBool(Env::printMightabilities))
         printf("Optimization not possible: current trajectory contains one or zero local path\n");
    return 1;
  }

  if(rand){
    ChronoOn();
    double tua = 0, toot = 0;
    while (fabs(gaintot - 1.0) < EPS6 && nb_optim < MAX_NB_TRY_OPTIM) {
      for (i = 1; i <= p3d_get_NB_OPTIM() /*&& (gain > 0.0001 || gain == 0)*/; i++) {
        if (p3d_optim_traj(traj, &gain, &ntest)) {
          gaintot = gaintot * (1. - gain);
        }
        toot = tua;
//         if (fct_draw) {
//           (*fct_draw)();
//         }
        if (!fct_stop_optim()) {
          g3d_draw_allwin_active();
          if (ENV.getBool(Env::printMightabilities))
            ChronoPrint("");
          double tu = 0.0, ts = 0.0;
          ChronoTimes(&tu, &ts);
          if (getStatStatus()) {
            XYZ_GRAPH->stat->postTime += tu;
          }
          ChronoOff();
          return 0;
        }
      }
      nb_optim++;
    }
    if (fabs(gaintot - 1.0) > EPS6) {
      /* the curve has been optimized */
      p3d_simplify_traj(traj);
    }
    gaintot = (1. - gaintot) * 100.;
    if (ENV.getBool(Env::printMightabilities))
    {
        printf("The trajectory is optimised by %f%%\n", gaintot);
        printf("nb collision test : %d\n", ntest);
        ChronoPrint("");
    }
    double tu = 0.0, ts = 0.0;
    ChronoTimes(&tu, &ts);
    if (getStatStatus()) {
      XYZ_GRAPH->stat->postTime += tu;
    }
    ChronoOff();
  }

  if(elastic && p3d_get_ik_choice() == IK_NORMAL){
    epsilon = 0.001 * p3d_get_eps_elastic();
    printf("\ngradient descent optimization \n");
    p3d_gradientDescentOptimize(traj, 0, epsilon, p3d_get_d0(), p3d_get_QUICK_DESCENT(),fct_stop_optim, fct_draw);
  }else if (clean && p3d_get_ik_choice() == IK_NORMAL){
    if (ENV.getBool(Env::printMightabilities))
      printf("\nclearing trajectory \n");
    p3d_clearTraj(traj, 0, fct_stop_optim);
  }
  return 0;
}


/*
 *  One step of optimization:
 *      2 configurations are randomly chosen on the trajectory
 *      the local planner is called between these configurations
 *      If the local path obtained is collision free and shorter
 *      than the initial portion, it replaces the portion in the
 *      trajectory.
 */

int p3d_optim_traj(p3d_traj *trajPt, double *gain, int *ntest) {
  p3d_rob *robotPt = trajPt->rob;
  p3d_localpath *localpath1Pt = NULL, *localpath2Pt = NULL,
                                *localpath3Pt = NULL, *cur_localpathPt = NULL;
  p3d_localpath *opt_pathPt[3] = {NULL, NULL, NULL};
  p3d_localpath *end_localpath_qiq1Pt, *start_localpath_q1q2Pt,
  *end_localpath_q1q2Pt = NULL, *start_localpath_q2qePt = NULL;
  p3d_localpath *start_new_trajPt = NULL,
                                    *end_new_trajPt = NULL;

  int q2_found = FALSE;
  double l = 0, l1 = 0, l2 = 0, ltot;
  double loc_dist1, lcur = 0, lnext = 0;
  configPt q1 = NULL, q2 = NULL, qinit = NULL, qgoal = NULL;
  double opt_cost[3], init_cost[3];
  double cost[8], cost_tmp[8], cost_min;
  int sort[8], collision[4] = { -1, -1, -1};
  int i, j, k, jmin = 0;
  char opt[3];
  int path_found = FALSE;
  int validIkSol[3] = {1,1,1};
  int flag = 0;
  /* for debugging */
  int i1 = 0, i2 = 0;
  double total_cost = 0.0;

  init_cost[0] = 0;
  *gain = 0.0;

  /* maximal penetration distance allowed */
  /* if trajectory is made of only one local path, no smoothing */
  if (trajPt->nlp < 2)
    return FALSE;

  /* length of trajPt */
  ltot = p3d_ends_and_length_traj(trajPt, &qinit, &qgoal);
  if (ltot <= 3*EPS6) {
    /* trajectory too short */
    p3d_destroy_config(robotPt, qinit);
    p3d_destroy_config(robotPt, qgoal);
    return FALSE;
  }

  /* pick random pairs of configurations. Make sure 0 < l1 < l2 < ltot */
  l2 = p3d_random(EPS6 * ltot, (1 - EPS6) * ltot);
  l1 = p3d_random(.25 * EPS6 * ltot, l2 - .25 * EPS6 * ltot);

  /* look for the first local path containing q1 */
  localpath1Pt = trajPt->courbePt;
  l = localpath1Pt->length_lp;


  while ((l < l1) && (localpath1Pt != NULL)) {
    init_cost[0] += localpath1Pt->cost(robotPt, localpath1Pt);
    localpath1Pt = localpath1Pt->next_lp;
    i1++;
    l += localpath1Pt->length_lp;
  }



  l -= localpath1Pt->length_lp;
  loc_dist1 = l1 - l;
  /* the local path containing q1 has been found */
  /* the portion of the local path containing q1 up to q1 is extracted
     in order to compute its cost */
  end_localpath_qiq1Pt =
    localpath1Pt->extract_sub_localpath(robotPt, localpath1Pt,
                                        0, loc_dist1);
  init_cost[0] += end_localpath_qiq1Pt->cost(robotPt, end_localpath_qiq1Pt);

  if(!p3d_compare_iksol(robotPt->cntrt_manager, trajPt->courbePt->ikSol, localpath1Pt->ikSol)){
    validIkSol[0] = 0;
  }

  localpath2Pt = localpath1Pt;

  if (l2 - l1 > localpath1Pt->length_lp - loc_dist1) {
    /* q2 is not on the same local path as q1 */
    /* get the last part of the localpath containing q1*/
    start_localpath_q1q2Pt =
      localpath1Pt->extract_sub_localpath(robotPt,
                                          localpath1Pt, loc_dist1,
                                          localpath1Pt->length_lp);
    lcur = l1 - loc_dist1 + localpath1Pt->length_lp; // l + localpath1Pt->length_lp
  } else {
    /* q1 and q2 are on the same local path */
    start_localpath_q1q2Pt =
      localpath1Pt->extract_sub_localpath(robotPt,
                                          localpath1Pt, loc_dist1,
                                          l2 - l1 + loc_dist1);
    /* compute second configuration */
    q2 = localpath1Pt->config_at_distance(robotPt,
                                          localpath1Pt,
                                          l2 - l1 + loc_dist1);
    if (robotPt->cntrt_manager->cntrts != NULL) {
      p3d_set_and_update_this_robot_conf(robotPt, q2);
      p3d_get_robot_config_into(robotPt, &q2);
    }

    /* copy second part of the path */
    end_localpath_q1q2Pt = start_localpath_q1q2Pt;
    /* copy beginning of thirs part of the path */
    start_localpath_q2qePt =
      localpath1Pt->extract_sub_localpath(robotPt, localpath1Pt,
                                          l2 - l1 + loc_dist1,
                                          localpath1Pt->length_lp);
    q2_found = TRUE;
  }
  /* compute the first configuration */
  q1 = start_localpath_q1q2Pt->
       config_at_distance(robotPt, start_localpath_q1q2Pt, 0);
  if (robotPt->cntrt_manager->cntrts != NULL) {
    p3d_set_and_update_this_robot_conf_multisol(robotPt, q1,NULL, 0, localpath1Pt->ikSol);
    p3d_get_robot_config_into(robotPt, &q1);
    p3d_set_robot_iksol(robotPt, robotPt->ikSol);
  }

  /* cost of the second part of the trajectory (between q1 and q2)
     initialized */
  init_cost[1] =
    start_localpath_q1q2Pt->cost(robotPt, start_localpath_q1q2Pt);

  i2 = i1;
  while (q2_found == FALSE) {
    i2++;
    localpath2Pt = localpath2Pt->next_lp;
    lnext = lcur + localpath2Pt->length_lp;


    if (lnext < l2) {
      init_cost[1] += localpath2Pt->cost(robotPt, localpath2Pt);
    } else {
      /* the local path containing q2 has been found */
      q2_found = TRUE;
      end_localpath_q1q2Pt =
        localpath2Pt->extract_sub_localpath(robotPt,
                                            localpath2Pt, 0,
                                            l2 - lcur);
      start_localpath_q2qePt =
        localpath2Pt->extract_sub_localpath(robotPt, localpath2Pt,
                                            l2 - lcur,
                                            localpath2Pt->length_lp);

      /* update cost of second and third sections */
      init_cost[1] +=
        end_localpath_q1q2Pt->cost(robotPt, end_localpath_q1q2Pt);
      q2 = localpath2Pt->config_at_distance(robotPt, localpath2Pt,
                                            l2 - lcur);
      if (robotPt->cntrt_manager->cntrts != NULL) {
          p3d_set_and_update_this_robot_conf_multisol(robotPt, q2, NULL, 0, localpath1Pt->ikSol);
          p3d_get_robot_config_into(robotPt, &q2);
          p3d_set_robot_iksol(robotPt, robotPt->ikSol);
        }
    }
    lcur = lnext;
  }


  if(!p3d_compare_iksol(robotPt->cntrt_manager, localpath1Pt->ikSol, localpath2Pt->ikSol)){
    validIkSol[1] = 0;
  }
  /* cost of the last part of the initial path */
  init_cost[2] =
    start_localpath_q2qePt->cost(robotPt, start_localpath_q2qePt);

  localpath3Pt = localpath2Pt->next_lp;

  while (localpath3Pt != NULL) {
    init_cost[2] += localpath3Pt->cost(robotPt, localpath3Pt);
    if(localpath3Pt->next_lp == NULL){//last
      if(!p3d_compare_iksol(robotPt->cntrt_manager, localpath2Pt->ikSol, localpath3Pt->ikSol)){
        validIkSol[2] = 0;
      }
    }
    localpath3Pt = localpath3Pt->next_lp;
  }
  /* At this point:
     - end_localpath_qiq1Pt is a copy of the portion of the local path
     of the initial trajectory containing q1 up to q1
     - start_localpath_q1q2Pt is a copy of the portion of the local
     path of the initial trajectory containing q1 starting at q1

     - end_localpath_q1q2Pt is a copy of the portion of the local path
     of the initial trajectory containing q2 up to q2
     - start_localpath_q2qePt is a copy of the portion of the local
     path of the initial trajectory containing q2 starting at q2

     - localpath1Pt points toward the local path of the initial
     trajectory containing q1
     - localpath2Pt points toward the local path of the initial
     trajectory containing q2

     - init_cost[0], init_cost[1], init_cost[2] are the respective costs of the 3 portion
     of trajectory (qi,q1) (q1,q2) (q2,qe)
  */

  /* the local planner is called between qi and q1, q1 and q2, q2 and qe*/
  opt_pathPt[0] = p3d_local_planner(robotPt, qinit, q1);
  p3d_copy_iksol(robotPt->cntrt_manager, localpath1Pt->ikSol, &((opt_pathPt[0])->ikSol));

  opt_pathPt[1] = p3d_local_planner(robotPt, q1, q2);
  p3d_copy_iksol(robotPt->cntrt_manager, localpath1Pt->ikSol, &((opt_pathPt[1])->ikSol));
  opt_pathPt[2] = p3d_local_planner(robotPt, q2, qgoal);
  p3d_copy_iksol(robotPt->cntrt_manager, localpath2Pt->ikSol, &((opt_pathPt[2])->ikSol));

  p3d_destroy_config(robotPt, q1);
  p3d_destroy_config(robotPt, q2);
  p3d_destroy_config(robotPt, qinit);
  p3d_destroy_config(robotPt, qgoal);

  for (i = 0; i <= 2; i++) {
    if (opt_pathPt[i] == NULL) {
      opt_cost[i] = P3D_HUGE;
    } else {
      opt_cost[i] = opt_pathPt[i]->cost(robotPt, opt_pathPt[i]);
    }
  }

  /* collision checking is performed as late as possible */
  /* there are 8 possible trajectories: we compute the cost for each
     of them and store it in cost[i]. We then sort the trajectories
     by increasing cost. The cheapest one without collision is
     chosen */

  /* cost of each trajectory */
  /* loop over the 8 different trajectories */
  for (i = 0; i <= 7; i++) {
    opt[0] = i & 1;     /* whether first part replaced or not */
    opt[1] = i & 2;     /* whether second part replaced or not */
    opt[2] = i & 4;     /* whether third part replaced or not */
    cost[i] = 0;

    /* loop over the three portions of the path */
    for (j = 0; j <= 2; j++) {
      if (opt[j] != 0) {
        cost[i] += opt_cost[j];
      } else {
        cost[i] += init_cost[j];
      }
      cost_tmp[i] = cost[i];
    }
  }
  /* sort the different trajectories by cost */
  for (i = 0; i <= 7; i++) {
    cost_min = P3D_HUGE;
    for (j = 0; j <= 7; j++) {
      /* look for the smallest cost */
      if (cost_tmp[j] < cost_min) {
        cost_min = cost_tmp[j];
        jmin = j;
      }
    }
    sort[i] = jmin;
    cost_tmp[jmin] = 2 * P3D_HUGE;
  }


  i = 0;
  while (!path_found) {
    /* best path */
    j = sort[i];
    opt[0] = j & 1;
    opt[1] = j & 2;
    opt[2] = j & 4;

    /* collision checking of the new local paths involved */
    k = 0;
    collision[3] = FALSE;
    flag = 0;
    while ((k <= 2) && (collision[3] == FALSE)) {
      /* is the k-th portion of this trajectory the new local path ?*/
      if(!validIkSol[k]){
        flag = 1;
      }
      if (opt[k] != 0) {
        /* if yes, has collision already been tested for the new
           local path ?*/
        if (collision[k] == -1) {
          if (p3d_unvalid_localpath_test(robotPt, opt_pathPt[k], ntest) || flag == 1 ) {            // <- modif Xav
         // if (p3d_col_test_localpath(robotPt, opt_pathPt[k], ntest) || flag == 1 ) {
            collision[k] = 1;
          } else {
            collision[k] = 0;
          }
        }
        if (collision[k] == 1) {
          collision[3] = TRUE;
        }
      }
      k++;
      flag = 0;
    }
    if (collision[3] == FALSE) {
      /* all the new local paths in this trajectory are collision free */
      path_found = TRUE;
    } else {
      /* otherwise check next best trajectory */
      i++;
    }
  }
  /* The best trajectory without collision has been found. We need
     now to concatenate the different parts */
  /* PrintInfo(("  Path chosen: %d %d %d \n", opt[0], opt[1], opt[2])); */

  /* If the best trajectory is the initial one, just return the same */
  if (((opt[0] == 0) && (opt[1] == 0) && (opt[2] == 0))) {
    /* destroy the local paths allocated in this function */
    end_localpath_qiq1Pt->destroy(robotPt, end_localpath_qiq1Pt);
    end_localpath_qiq1Pt = NULL;
    start_localpath_q1q2Pt->destroy(robotPt, start_localpath_q1q2Pt);
    if (end_localpath_q1q2Pt != start_localpath_q1q2Pt) {
      end_localpath_q1q2Pt->destroy(robotPt, end_localpath_q1q2Pt);
    }
    start_localpath_q2qePt->destroy(robotPt, start_localpath_q2qePt);
    start_localpath_q1q2Pt = NULL;
    end_localpath_q1q2Pt = NULL;
    start_localpath_q2qePt = NULL;

    if (opt_pathPt[0] != NULL) {
      opt_pathPt[0]->destroy(robotPt, opt_pathPt[0]);
    }
    if (opt_pathPt[1] != NULL) {
      opt_pathPt[1]->destroy(robotPt, opt_pathPt[1]);
    }
    if (opt_pathPt[2] != NULL) {
      opt_pathPt[2]->destroy(robotPt, opt_pathPt[2]);
    }
    return FALSE;
  }

  /* first portion of the new path */
  if (opt[0] == 0) {
    /*destroy useless local path */
    if (opt_pathPt[0] != NULL) {
      opt_pathPt[0]->destroy(robotPt, opt_pathPt[0]);
      opt_pathPt[0] = NULL;
    }
    /* first portion not replaced */
    cur_localpathPt = trajPt->courbePt;
    while (cur_localpathPt != localpath1Pt) {
      if (start_new_trajPt == NULL) {
        start_new_trajPt = cur_localpathPt->copy(robotPt, cur_localpathPt);
        end_new_trajPt = start_new_trajPt;
      } else {
        end_new_trajPt =
          append_to_localpath(end_new_trajPt,
                              cur_localpathPt->copy(robotPt,
                                                    cur_localpathPt));
      }
      cur_localpathPt = cur_localpathPt->next_lp;
    }

    if (start_new_trajPt == NULL) {
      /* q1 is on the first local path */
      start_new_trajPt = end_localpath_qiq1Pt;
      end_new_trajPt = start_new_trajPt;
    } else {
      end_new_trajPt = append_to_localpath(end_new_trajPt,
                                           end_localpath_qiq1Pt);
    }
  } else {
    /* first local path replaced */
    start_new_trajPt = opt_pathPt[0];
    end_new_trajPt = start_new_trajPt;
    /* destroy useless local path */
    end_localpath_qiq1Pt->destroy(robotPt, end_localpath_qiq1Pt);
    end_localpath_qiq1Pt = NULL;
  }

  /* second portion of the new path */
  if (opt[1] == 0) {
    /* second local path not replaced */
    /* add copy of beginning of second part */
    end_new_trajPt = append_to_localpath(end_new_trajPt,
                                         start_localpath_q1q2Pt);

    /*destroy useless local path */
    if (opt_pathPt[1] != NULL) {
      opt_pathPt[1]->destroy(robotPt, opt_pathPt[1]);
      opt_pathPt[1] = NULL;
    }

    if (localpath2Pt != localpath1Pt) {
      /* q1 and q2 are not on the same local path */
      cur_localpathPt = localpath1Pt->next_lp;
      while (cur_localpathPt != localpath2Pt) {
        end_new_trajPt =
          append_to_localpath(end_new_trajPt,
                              cur_localpathPt->copy(robotPt,
                                                    cur_localpathPt));
        cur_localpathPt = cur_localpathPt->next_lp;
      }
      end_new_trajPt = append_to_localpath(end_new_trajPt,
                                           end_localpath_q1q2Pt);
    }
  } else {
    /* second local path replaced */
    end_new_trajPt = append_to_localpath(end_new_trajPt,
                                         opt_pathPt[1]);
    /* destroy useless local paths */
    start_localpath_q1q2Pt->destroy(robotPt, start_localpath_q1q2Pt);
    if (end_localpath_q1q2Pt != start_localpath_q1q2Pt) {
      end_localpath_q1q2Pt->destroy(robotPt, end_localpath_q1q2Pt);
    }
    start_localpath_q1q2Pt = NULL;
    end_localpath_q1q2Pt = NULL;
  }

  /* third portion of the new path */
  if (opt[2] == 0) {
    /* third local path not replaced */
    end_new_trajPt = append_to_localpath(end_new_trajPt,
                                         start_localpath_q2qePt);

    /*destroy useless local path */
    if (opt_pathPt[2] != NULL) {
      opt_pathPt[2]->destroy(robotPt, opt_pathPt[2]);
      opt_pathPt[2] = NULL;
    }

    cur_localpathPt = localpath2Pt->next_lp;
    while (cur_localpathPt != NULL) {
      end_new_trajPt =
        append_to_localpath(end_new_trajPt,
                            cur_localpathPt->copy(robotPt, cur_localpathPt));
      cur_localpathPt = cur_localpathPt->next_lp;
    }
  } else {
    /* second local path replaced */
    end_new_trajPt = append_to_localpath(end_new_trajPt, opt_pathPt[2]);

    /* destroy useless local paths */
    start_localpath_q2qePt->destroy(robotPt, start_localpath_q2qePt);
    start_localpath_q2qePt = NULL;
  }

  /* If two consecutive portions have not been replaced by the
     corresponding local paths, the local linking them has been cut
     in two pieces. We need to fix that a posteriori */



  /* destroy the initial trajectory and replace it by the new one */
  destroy_list_localpath(robotPt, trajPt->courbePt);
  /* replace it by the new one */
  trajPt->courbePt = start_new_trajPt;
  /* update the number of local paths */
  trajPt->nlp = p3d_compute_traj_nloc(trajPt);
  /* store the parameter range of this trajectory */
  trajPt->range_param = p3d_compute_traj_rangeparam(trajPt);

  total_cost = init_cost[0] + init_cost[1] + init_cost[2];
  *gain = (total_cost - cost[j]) / total_cost;

  return TRUE;
}

/*
 *  p3d_simplify_traj
 *
 *  Simplify a path  by removing parts by which the robot passes several
 *  time. This situation occurs only for Reeds and Shepp paths.
 *
 *  Description:
 *     For each local path composing the path, a specific function is called.
 *     This function removes the intersection between the local path given
 *     as input and the next local path.
 *
 */

void p3d_simplify_traj(p3d_traj *trajPt) {
  p3d_rob *robotPt = trajPt->rob;
  p3d_localpath *localpathPt = trajPt->courbePt, *new_localpathPt = NULL,
                               *nextlpPt = NULL, *prevlpPt = NULL;
  int ntest = 0;

  int need_colcheck = FALSE, replace = FALSE;

  if (localpathPt == NULL)
    return;

  while (localpathPt->next_lp != NULL) {
    replace = FALSE;
    new_localpathPt =
      localpathPt->simplify(robotPt, localpathPt, &need_colcheck);

    /* replace old part by new one only if they are different */
    if (new_localpathPt != localpathPt) {
      if (need_colcheck) {
        /* replace the old part by the new one only if no collision */
        if (!p3d_col_test_traj(robotPt, new_localpathPt, &ntest)) {
          replace = TRUE;
        }
      } else {
        replace = TRUE;
      }
    }

    if (replace) {
      /* replace localpathPt and localpathPt->next_lp by new_localpathPt
      and new_localpathPt->next_lp */

      /* keep a pointer on the rest of the path before and after the
      part to remove */
      nextlpPt = localpathPt->next_lp->next_lp;
      prevlpPt = localpathPt->prev_lp;
      /* destroy first localpathPt->next_lp */
      localpathPt->next_lp->destroy(robotPt, localpathPt->next_lp);
      /* and then destroy localpathPt */
      localpathPt->destroy(robotPt, localpathPt);
      localpathPt = new_localpathPt;

      /* insert the new part into the path */
      if (prevlpPt != NULL) {
        /* we are not at the beginning of the path */
        prevlpPt->next_lp = new_localpathPt;
        new_localpathPt->prev_lp = prevlpPt;
      } else {
        trajPt->courbePt = new_localpathPt;
        new_localpathPt->prev_lp = NULL;
      }
      if (new_localpathPt->next_lp != NULL) {
        /* go to second local path of new part */
        new_localpathPt = new_localpathPt->next_lp;
      }
      new_localpathPt->next_lp = nextlpPt;

      if (nextlpPt != NULL) {
        /* we are not at the end of the path */
        nextlpPt->prev_lp = new_localpathPt;
      }
    }

    if (localpathPt->next_lp != NULL) {
      /* usually localpathPt->next_lp is not NULL, but
         p3d_simplify_manh() can delete "empty" localpaths
         in the trajectory and at the end it might happen
         that localpathPt->next_lp no longer points to a
         localpath after simplifying */
      localpathPt = localpathPt->next_lp;
    }
  }
  /* update the number of local paths */
  trajPt->nlp = p3d_compute_traj_nloc(trajPt);
  trajPt->range_param = p3d_compute_traj_rangeparam(trajPt);
}

int fct_stop_optim(void) {
  double ts, tu, tmax = p3d_get_optimization_time();
  ChronoMicroTimes(&tu, &ts);
  if (p3d_get_use_optimization_time() && tu >= tmax) {
    printf("Optimization stoped by counter.\n");
    return FALSE;
  }
#ifdef LIGHT_PLANNER  
  if (hasReachedTimeLimit()) {
    printf("Optimization has reached time limit.\n");
    return FALSE;
  }
#endif
  if (p3d_GetStopValue()) {
    printf("Optimization stoped by user.\n");
    return FALSE;
  }
  if (STOP_OPTIM) {
    STOP_OPTIM = FALSE;
    printf("Optimization stoped by user.\n");
    return (FALSE);
  } else {
    return (TRUE);
  }
}

double p3d_get_d0() {
  return D0;
}

double p3d_get_eps_elastic() {
  return EPS_ELASTIC;
}

int p3d_get_QUICK_DESCENT() {
  return QUICK_DESCENT;
}

int print_state(int state,int nbit,int maxit,double time,double maxtime,double gain,double maxgain){
    int cand_state;
    if(maxit>0){
        cand_state=(int)floor(100*(double)nbit/(double)maxit);
        if(cand_state>state) state=cand_state;
    }
    if(maxtime>0){
        cand_state=(int)floor(100*time/maxtime);
        if(cand_state>state) state=cand_state;
    }
    if(maxgain>0){
        cand_state=(int)floor(100*gain/maxgain);
        if(cand_state>state) state=cand_state;
    }
    if(state>100) state=100;
    printf("\r     \r");
    if(state<100) printf(" ");
    if(state<10) printf(" ");
    printf("%d %c",state,'%');
    fflush(stdout);
    return state;
}

int p3d_simple_shorcut(p3d_traj *trajPt,double *gain,double *cost){

    //printf("START simple_shorcut\n");

    double u1,u2,u,old_cost,new_cost,cost_ini,c,old_range_param;
    int n,old_nlp;
    localpath *lp_cur,*lpI,*lpF,*lp1,*lp2,*lp3;
    p3d_rob *robotPt;
    configPt qI,qA1,qA2,qB2,qB3,qF;

    *gain=0;
    *cost=0;

    if(!trajPt || !trajPt->courbePt || !trajPt->rob) return FALSE;
    robotPt=trajPt->rob;

    // Selection of two random parameters u1 and u2 along the trajectory (u1<=u2)
    u1=p3d_random(0,trajPt->range_param);
    u2=p3d_random(0,trajPt->range_param);
    if(u2<u1){
        u=u2;
        u2=u1;
        u1=u;
    }

    // Research of lpI : the localpath corresponding to parameter u1
    // cost_ini : cost of the trajectory from trajPt->courbePt to lpI->prev_lp
    u=0;
    cost_ini=0;
    lp_cur=trajPt->courbePt;
    while(lp_cur && u1>u+lp_cur->range_param){
        u+=lp_cur->range_param;
        cost_ini+=lp_cur->cost(robotPt,lp_cur);
        lp_cur=lp_cur->next_lp;
    }
    if(!lp_cur){
        //printf("lpI not found\n");
        return FALSE;
    }
    lpI=lp_cur;

    // ALLOCATION qI qA1 lp1
    qI=lpI->config_at_param(robotPt,lpI,0);
    qA1=lpI->config_at_param(robotPt,lpI,u1-u);
    lp1=p3d_local_planner(robotPt,qI,qA1);

    // Testing lp1 validity
    if(!lp1 || p3d_unvalid_localpath_test(robotPt,lp1,&n)){
        // FREE qI qAI lp1;
        if(lp1){
            lp1->destroy(robotPt,lp1);
        }
        else{
            if(qI) p3d_destroy_config(robotPt,qI);
            if(qA1) p3d_destroy_config(robotPt,qA1);
        }
        //printf("lp1 is not valid\n");
        return FALSE;
    }

    old_cost=0;
    old_range_param=0;
    old_nlp=0;

    // Research of lpI : the localpath corresponding to parameter u2
    // cost_ini : cost of the trajectory from trajPt->courbePt to lpF->prev_lp
    // old_cost : cost of the trajectory from lpI to lpF->prev_lp
    // range_param : range of parameter from lpI to lpF->prev_lp
    // old_nlp : number of localpaths from lpI to lpF->prev_lp
    while(lp_cur && u2>u+lp_cur->range_param){
        u+=lp_cur->range_param;
        c=lp_cur->cost(robotPt,lp_cur);
        old_cost+=c;
        old_range_param+=lp_cur->range_param;
        cost_ini+=c;
        old_nlp++;
        lp_cur=lp_cur->next_lp;
    }
    if(!lp_cur){
        // FREE qI qA1 lp1
        lp1->destroy(robotPt,lp1);
        //printf("lpF not found\n");
        return FALSE;
    }
    lpF=lp_cur;

    // Update of cost_ini, old_cost, old_range_param and old_nlp to take lpF into account
    c=lpF->cost(robotPt,lpF);
    old_cost+=c;
    cost_ini+=c;
    old_range_param+=lpF->range_param;
    old_nlp++;

    // ALLOCATION qA2 qB2 lp2
    qA2=p3d_copy_config(robotPt,qA1);
    qB2=lpF->config_at_param(robotPt,lpF,u2-u);
    lp2=p3d_local_planner(robotPt,qA2,qB2);

    // Testing lp2 validity
    if(!lp2 || p3d_unvalid_localpath_test(robotPt,lp2,&n)){
        // FREE qI qA1 lp1
        if(lp1){
            lp1->destroy(robotPt,lp1);
        }
        else{
            if(qI) p3d_destroy_config(robotPt,qI);
            if(qA1) p3d_destroy_config(robotPt,qA1);
        }
        // FREE qA2 qB2 lp2
        if(lp2){
            lp2->destroy(robotPt,lp2);
        }
        else{
            if(qA2) p3d_destroy_config(robotPt,qA2);
            if(qB2) p3d_destroy_config(robotPt,qB2);
        }
        //printf("lp2 is not valid\n");
        /*
        if(lp2->type_lp==LINEAR){
            print_lin(robotPt,lp2);
        }
        if(lp2->type_lp==QUADSPLINE){
            print_quadspline(robotPt,lp2);
        }
        */
        return FALSE;
    }

    // ALLOCATION qB3 qF lp3
    qB3=p3d_copy_config(robotPt,qB2);
    qF=lpF->config_at_param(robotPt,lpF,lpF->range_param);
    lp3=p3d_local_planner(robotPt,qB3,qF);

    // Testing lp3 validity
    if(!lp3 || p3d_unvalid_localpath_test(robotPt,lp3,&n)){
        // FREE qI qA1 lp1
        if(lp1){
            lp1->destroy(robotPt,lp1);
        }
        else{
            if(qI) p3d_destroy_config(robotPt,qI);
            if(qA1) p3d_destroy_config(robotPt,qA1);
        }
        // FREE qA2 qB2 lp2
        if(lp2){
            lp2->destroy(robotPt,lp2);
        }
        else{
            if(qA2) p3d_destroy_config(robotPt,qA2);
            if(qB2) p3d_destroy_config(robotPt,qB2);
        }
        // FREE qB3 qF lp3
        if(lp3){
            lp3->destroy(robotPt,lp3);
        }
        else{
            if(qB3) p3d_destroy_config(robotPt,qB3);
            if(qF) p3d_destroy_config(robotPt,qF);
        }
        //printf("lp3 is not valid\n");
        return FALSE;
    }

    // new_cost : cost of lp1->lp2->lp3
    new_cost=lp1->cost(robotPt,lp1)+lp2->cost(robotPt,lp2)+lp3->cost(robotPt,lp3);

    // Testing for cost improvement
    if(new_cost>=old_cost){
        // FREE qI qA1 lp1
        if(lp1){
            lp1->destroy(robotPt,lp1);
        }
        else{
            if(qI) p3d_destroy_config(robotPt,qI);
            if(qA1) p3d_destroy_config(robotPt,qA1);
        }
        // FREE qA2 qB2 lp2
        if(lp2){
            lp2->destroy(robotPt,lp2);
        }
        else{
            if(qA2) p3d_destroy_config(robotPt,qA2);
            if(qB2) p3d_destroy_config(robotPt,qB2);
        }
        // FREE qB3 qF lp3
        if(lp3){
            lp3->destroy(robotPt,lp3);
        }
        else{
            if(qB3) p3d_destroy_config(robotPt,qB3);
            if(qF) p3d_destroy_config(robotPt,qF);
        }
        //printf("new_cost>=old_cost : new_cost=%f old_cost=%f\n",new_cost,old_cost);
        return FALSE;
    }

    // cost_ini for the rest of the trajectory
    lp_cur=lpF->next_lp;
    while(lp_cur){
        cost_ini+=lp_cur->cost(robotPt,lp_cur);
        lp_cur=lp_cur->next_lp;
    }

    // Updates for the new trajectory
    trajPt->nlp=trajPt->nlp-old_nlp+3;
    trajPt->range_param=trajPt->range_param-old_range_param+lp1->range_param+lp2->range_param+lp3->range_param;


    // Connections for the new trajectory
    if(!lpI->prev_lp){
        trajPt->courbePt=lp1;
    }
    else{
        lpI->prev_lp->next_lp=lp1;
    }
    lp1->prev_lp=lpI->prev_lp;
    lp1->next_lp=lp2;
    lp2->prev_lp=lp1;
    lp2->next_lp=lp3;
    lp3->prev_lp=lp2;
    lp3->next_lp=lpF->next_lp;
    if(lpF->next_lp) lpF->next_lp->prev_lp=lp3;

    // FREE lpI->lpF
    lp_cur=lpI;
    n=0;
    while(n<old_nlp){
        lpI=lp_cur->next_lp;
        lp_cur->destroy(robotPt,lp_cur);
        lp_cur=lpI;
        n++;
    }

    // cost and gain
    *cost=cost_ini-old_cost+new_cost;
    *gain=(old_cost-new_cost)/cost_ini;

    return TRUE;
}

double p3d_run_simple_shortcut(p3d_traj *trajPt,int maxit,double maxtime,double maxgain){
    if(maxit<0 && maxtime<0 && maxgain<0) return 0;
    FILE *f=fopen("load_random_shorcut_data.m","w");
    double time=0;
    int nbit=0,state=0;
    double gain=0,gain_one_loop;
    double cost_ini=p3d_compute_traj_cost(trajPt);
    double cost_after,cost_before;
    fprintf(f,"%c Random shorcut data generated by move3d\n\n",'%');
    fprintf(f,"DATA=[0 0 %f\n",cost_ini);
    ChronoTimeOfDayOn();
    while((maxit<0 || nbit<maxit) && (maxtime<0 || time<maxtime) && (maxgain<0 || gain<maxgain)){
        cost_before=p3d_compute_traj_cost(trajPt);
        p3d_simple_shorcut(trajPt,&gain_one_loop,&cost_after);
        nbit++;
        //printf("%d\n",nbit);
        ChronoTimeOfDayTimes(&time);
        cost_after=p3d_compute_traj_cost(trajPt);
        gain=1-cost_after/cost_ini;
        fprintf(f,"%f %f %f\n",time,gain,cost_after);
        state=print_state(state,nbit,maxit,time,maxtime,gain,maxgain);
    }
    fprintf(f,"];\n\n");
    fprintf(f,"save DATA;");
    ChronoTimeOfDayOff();
    fclose(f);
    printf("\n");
    return gain;
}

void p3d_perturb_quad_conf(p3d_rob *robotPt,configPt q,double dq){

    double norm,dvx,dvy,dvz,dax,day,daz;
    lm_quadspline_str *paramPt=lm_get_quadspline_lm_param(robotPt);

    // MALLOC q_dir
    configPt q_dir=p3d_alloc_config(robotPt);
    norm=0;
    while(norm<EPS4){
        p3d_shoot(robotPt,q_dir,0);
        dvx=q_dir[paramPt->vx_dof]-q[paramPt->vx_dof];
        dvy=q_dir[paramPt->vy_dof]-q[paramPt->vy_dof];
        dvz=q_dir[paramPt->vz_dof]-q[paramPt->vz_dof];
        dax=q_dir[paramPt->ax_dof]-q[paramPt->ax_dof];
        day=q_dir[paramPt->ay_dof]-q[paramPt->ay_dof];
        daz=q_dir[paramPt->az_dof]-q[paramPt->az_dof];
        norm=sqrt(dvx*dvx+dvy*dvy+dvz*dvz+dax*dax+day*day+daz*daz);
    }
    // FREE q_dir
    p3d_destroy_config(robotPt,q_dir);

    dvx=dq*dvx/norm;
    dvy=dq*dvy/norm;
    dvz=dq*dvz/norm;
    dax=dq*dax/norm;
    day=dq*day/norm;
    daz=dq*daz/norm;

    q[paramPt->vx_dof]=q[paramPt->vx_dof]+dvx;
    q[paramPt->vy_dof]=q[paramPt->vy_dof]+dvy;
    q[paramPt->vz_dof]=q[paramPt->vz_dof]+dvz;
    q[paramPt->ax_dof]=q[paramPt->ax_dof]+dax;
    q[paramPt->ay_dof]=q[paramPt->ay_dof]+day;
    q[paramPt->az_dof]=q[paramPt->az_dof]+daz;

    if(q[paramPt->vx_dof]>paramPt->vmax) q[paramPt->vx_dof]=paramPt->vmax;
    if(q[paramPt->vx_dof]<-paramPt->vmax) q[paramPt->vx_dof]=-paramPt->vmax;
    if(q[paramPt->vy_dof]>paramPt->vmax) q[paramPt->vy_dof]=paramPt->vmax;
    if(q[paramPt->vy_dof]<-paramPt->vmax) q[paramPt->vy_dof]=-paramPt->vmax;
    if(q[paramPt->vz_dof]>paramPt->vmax) q[paramPt->vz_dof]=paramPt->vmax;
    if(q[paramPt->vz_dof]<-paramPt->vmax) q[paramPt->vz_dof]=-paramPt->vmax;

    if(q[paramPt->ax_dof]>paramPt->amax) q[paramPt->ax_dof]=paramPt->amax;
    if(q[paramPt->ax_dof]<-paramPt->amax) q[paramPt->ax_dof]=-paramPt->amax;
    if(q[paramPt->ay_dof]>paramPt->amax) q[paramPt->ay_dof]=paramPt->amax;
    if(q[paramPt->ay_dof]<-paramPt->amax) q[paramPt->ay_dof]=-paramPt->amax;
    if(q[paramPt->az_dof]>paramPt->amax) q[paramPt->az_dof]=paramPt->amax;
    if(q[paramPt->az_dof]<-paramPt->amax) q[paramPt->az_dof]=-paramPt->amax;
}

p3d_localpath *p3d_one_step_random_gradient_descent(p3d_rob* robotPt,p3d_traj *trajPt,p3d_localpath *lp1,double dq){

    double cost_after,cost_before;
    p3d_localpath *lpa,*lpb,*lp2;
    configPt q,qn,q0,qF;
    int n;

    if(!robotPt || !trajPt || !lp1 || !lp1->next_lp){
        printf("error: invalid parameter(s) in p3d_one_step_random_grandient_descent\n");
        return NULL;
    }
    lp2=lp1->next_lp;

    if(robotPt->lpl_type!=QUADSPLINE || lp1->type_lp!=QUADSPLINE || lp2->type_lp!=QUADSPLINE){
        printf("error: localpath of type QUADSPLINE expected in p3d_one_step_random_grandient_descent\n");
        return NULL;
    }

    cost_before=lp1->cost(robotPt,lp1)+lp2->cost(robotPt,lp2);

    // MALLOC q0 q qF;
    q0=lp1->config_at_param(robotPt,lp1,0);
    q=lp2->config_at_param(robotPt,lp2,0);
    qF=lp2->config_at_param(robotPt,lp2,lp2->range_param);

    qn=p3d_copy_config(robotPt,q);
    p3d_perturb_quad_conf(robotPt,qn,dq);

    // MALLOC lpa lpb
    lpa=p3d_local_planner(robotPt,q0,qn);
    lpb=p3d_local_planner(robotPt,qn,qF);

    if(   !lpa || p3d_unvalid_localpath_test(robotPt,lpa,&n)
       || !lpb || p3d_unvalid_localpath_test(robotPt,lpb,&n)){
        // FREE lpa lpb qn
        lpa->destroy(robotPt,lpa);
        lpb->destroy(robotPt,lpb);
        p3d_destroy_config(robotPt,qn);

    }
    else{
        cost_after=lpa->cost(robotPt,lpa)+lpb->cost(robotPt,lpb);
        if(cost_after<cost_before){

            // Connections
            if(!lp1->prev_lp){
                trajPt->courbePt=lpa;
                lpa->prev_lp = NULL;
            }
            else{
                lp1->prev_lp->next_lp=lpa;
                lpa->prev_lp = lp1->prev_lp;
            }
            lpa->next_lp=lpb;
            lpb->prev_lp=lpa;
            if(lp2->next_lp){
                lp2->next_lp->prev_lp=lpb;
                lpb->next_lp = lp2->next_lp;
            }
            else
                //lpb->next_lp = NULL;


            // range param updates
            trajPt->range_param=(trajPt->range_param)-(lp1->range_param)+(lpa->range_param)
                                                     -(lp2->range_param)+(lpb->range_param);

            // FREE lp1 lp2 q
            lp1->destroy(robotPt,lp1);
            lp2->destroy(robotPt,lp2);
            p3d_destroy_config(robotPt,q);

            q=qn;
            lp1=lpa;
        }
        else{
            // FREE lpa lpb qn
            lpa->destroy(robotPt,lpa);
            lpb->destroy(robotPt,lpb);
            p3d_destroy_config(robotPt,qn);
        }
    }

    p3d_destroy_config(robotPt,q0);
    p3d_destroy_config(robotPt,q);
    p3d_destroy_config(robotPt,qF);

    return lp1;

}

double p3d_run_random_gradient_descent(p3d_rob *robotPt,p3d_traj *trajPt,p3d_localpath *lp,double dq,int maxit,double maxtime,double maxgain){

    double cost_ini,cost,time,gain,g;
    int nbit;

    if(maxit<0 && maxtime<0 && maxgain<0) return 0;

    if(!robotPt){
        printf("error: robot is NULL in p3d_run_random_grandient_descent\n");
        return 0;
    }

    if(!trajPt){
        printf("error: traj is NULL in p3d_run_random_grandient_descent\n");
        return 0;
    }

    if(!lp){
        printf("error: lp is NULL in p3d_run_random_grandient_descent\n");
        return 0;
    }

    if(!lp->next_lp){
        printf("error: lp->next_lp is NULL in p3d_run_random_grandient_descent\n");
        return 0;
    }

 /*
    if(p3d_unvalid_localpath_test(robotPt,lp,&nbit)){
        printf("error: lp is invalid in p3d_run_random_grandient_descent\n");
        return 0;
    }

    if(p3d_unvalid_localpath_test(robotPt,lp->next_lp,&nbit)){
        printf("error: lp->next_lp is invalid in p3d_run_random_grandient_descent\n");
        return 0;
    }
*/

    if(robotPt->lpl_type!=QUADSPLINE || lp->type_lp!=QUADSPLINE || lp->next_lp->type_lp!=QUADSPLINE){
        printf("error: localpath of type QUADSPLINE expected in p3d_run_random_grandient_descent\n");
        return 0;
    }

    cost_ini=lp->cost(robotPt,lp)+(lp->next_lp)->cost(robotPt,lp->next_lp);
    cost=cost_ini;
    time=0;
    nbit=0;
    gain=0;
    ChronoTimeOfDayOn();
    while(lp && (maxit<0 || nbit<maxit) && (maxtime<0 || time<maxtime) && (gain<0 || maxgain<0 || gain<maxgain)){
        lp=p3d_one_step_random_gradient_descent(robotPt,trajPt,lp,dq);
        if(lp) cost=lp->cost(robotPt,lp)+(lp->next_lp)->cost(robotPt,lp->next_lp);
        gain=1-cost/cost_ini;
        ChronoTimeOfDayTimes(&time);
        nbit++;
    }
    ChronoTimeOfDayOff();

    return gain;
}
